/* vim: set sw=8: -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * gog-renderer-pixbuf.c :
 *
 * Copyright (C) 2003 Jody Goldberg (jody@gnome.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#warning Uses deprecated pango ft2 api
#undef PANGO_DISABLE_DEPRECATED

#include <gnumeric-config.h>
#include <goffice/graph/gog-renderer-pixbuf.h>
#include <goffice/graph/gog-renderer-impl.h>
#include <goffice/graph/gog-style.h>
#include <goffice/graph/gog-view.h>
#include <goffice/utils/go-color.h>
#include <goffice/utils/go-font.h>
#include <goffice/utils/go-marker.h>
#include <goffice/utils/go-units.h>

#include <libart_lgpl/art_render_gradient.h>
#include <libart_lgpl/art_render_svp.h>
#include <libart_lgpl/art_render_mask.h>
#include <pango/pangoft2.h>
#include <gsf/gsf-impl-utils.h>
#include <src/application.h>

#include <math.h>

struct _GogRendererPixbuf {
	GogRenderer base;

	int w, h;
	GdkPixbuf *buffer;
	guchar    *pixels; /* from pixbuf */
	int	   rowstride;

	PangoContext *pango_context;
};

typedef GogRendererClass GogRendererPixbufClass;

static GObjectClass *parent_klass;

static void
gog_renderer_pixbuf_finalize (GObject *obj)
{
	GogRendererPixbuf *prend = GOG_RENDERER_PIXBUF (obj);

	if (prend->buffer != NULL) {
		g_object_unref (prend->buffer);
		prend->buffer = NULL;
	}

	if (prend->pango_context != NULL) {
		g_object_unref (prend->pango_context);
		prend->pango_context = NULL;
	}

	if (parent_klass != NULL && parent_klass->finalize != NULL)
		(parent_klass->finalize) (obj);
}

static void
gog_renderer_pixbuf_draw_path (GogRenderer *rend, ArtVpath const *path)
{
	GogRendererPixbuf *prend = GOG_RENDERER_PIXBUF (rend);
	GogStyle const *style = rend->cur_style;
	double width = gog_renderer_line_size (rend, style->line.width);
	ArtSVP *svp = art_svp_vpath_stroke ((ArtVpath *)path,
		ART_PATH_STROKE_JOIN_MITER, ART_PATH_STROKE_CAP_SQUARE,
		width, 4, 0.5);
	go_color_render_svp (style->line.color, svp,
		0, 0, prend->w, prend->h,
		prend->pixels, prend->rowstride);
	art_svp_free (svp);
}

static ArtRender *
gog_art_renderer_new (GogRendererPixbuf *prend)
{
	return art_render_new (0, 0, prend->w, prend->h,
		prend->pixels, prend->rowstride,
		gdk_pixbuf_get_n_channels (prend->buffer) - 1,
		8, ART_ALPHA_SEPARATE, NULL);
}

static void
gog_renderer_pixbuf_draw_polygon (GogRenderer *rend, ArtVpath const *path,
				  gboolean narrow)
{
	GogRendererPixbuf *prend = GOG_RENDERER_PIXBUF (rend);
	GogStyle const *style = rend->cur_style;
	ArtRender *render;
	ArtSVP *fill, *outline = NULL;
	ArtDRect bbox;
	ArtGradientLinear gradient;
	ArtGradientStop stops[2];
	GdkPixbuf *image;
	gint i, j, imax, jmax, w, h, x, y;

	if (!narrow && style->outline.width >= 0.)
		outline = art_svp_vpath_stroke ((ArtVpath *)path,
			ART_PATH_STROKE_JOIN_MITER, ART_PATH_STROKE_CAP_SQUARE,
			gog_renderer_line_size (rend, style->outline.width), 4, 0.5);

	if (style->fill.type != GOG_FILL_STYLE_NONE) {
		fill = art_svp_from_vpath ((ArtVpath *)path);
#if 0 /* art_svp_minus is not implemented */
		if (outline != NULL) {
			ArtSVP *tmp = art_svp_minus (fill, outline);
			art_svp_free (fill);
			fill = tmp;
		}
#endif

		switch (style->fill.type) {
		case GOG_FILL_STYLE_PATTERN:
			go_pattern_render_svp (&style->fill.u.pattern.pat,
				fill, 0, 0, prend->w, prend->h,
				prend->pixels, prend->rowstride);
			break;

		case GOG_FILL_STYLE_GRADIENT: {

			art_vpath_bbox_drect ((ArtVpath *)path, &bbox);
			render = gog_art_renderer_new (prend);
			art_render_svp (render, fill);
			
			go_gradient_setup (&gradient,
					   style->fill.u.gradient.dir,
					   style->fill.u.gradient.start, style->fill.u.gradient.end,
					   bbox.x0, bbox.y0, bbox.x1, bbox.y1,
					   stops);

			art_render_gradient_linear (render,
				&gradient, ART_FILTER_NEAREST);
			art_render_invoke (render);
			break;
		}

		case GOG_FILL_STYLE_IMAGE: {
			double dx, dy;

			image = style->fill.u.image.image;
			if (image == NULL)
				break;
			
			art_vpath_bbox_drect (path, &bbox);
			dx = bbox.x1 - bbox.x0;
			dy = bbox.y1 - bbox.y0;
			switch (style->fill.u.image.type) {
			case GOG_IMAGE_STRETCHED:
				gdk_pixbuf_composite (image, prend->buffer,
					bbox.x0, bbox.y0, dx, dy, bbox.x0, bbox.y0,
					dx / gdk_pixbuf_get_width (image),
					dy / gdk_pixbuf_get_height (image),
					GDK_INTERP_HYPER, 255);
				break;

			case GOG_IMAGE_WALLPAPER:
				imax = dx / (w = gdk_pixbuf_get_width (image));
				jmax = dy / (h = gdk_pixbuf_get_height (image));
				x = bbox.x0;
				for (i = 0; i < imax; i++) {
					y = bbox.y0;
					for (j = 0; j < jmax; j++) {
						gdk_pixbuf_copy_area (image, 0, 0, w, h,
								      prend->buffer, x, y);
						y += h;
					}
					gdk_pixbuf_copy_area (image, 0, 0, w,
							      prend->h % h, prend->buffer, x, y);
					x += w;
				}
				y = bbox.y0;
				for (j = 0; j < jmax; j++) {
					gdk_pixbuf_copy_area (image, 0, 0, (int)dx % w, h,
							      prend->buffer, x, y);
					y += h;
				}
				gdk_pixbuf_copy_area (image, 0, 0, (int)dx % w, (int)dy % h,
						      prend->buffer, x, y);
				break;
			}
			break;
		}

		case GOG_FILL_STYLE_NONE:
			break; /* impossible */
		}
		if (fill != NULL)
			art_svp_free (fill);
	}

	if (outline != NULL) {
		go_color_render_svp (style->outline.color, outline,
			0, 0, prend->w, prend->h,
			prend->pixels, prend->rowstride);
		art_svp_free (outline);
	}
}

static PangoLayout *
make_layout (GogRendererPixbuf *prend, char const *text)
{
	PangoLayout *layout;
	PangoAttribute *attr;
	PangoAttrList  *attrs = NULL;

	if (prend->pango_context == NULL)
		prend->pango_context = pango_ft2_get_context (
			application_display_dpi_get (TRUE),
			application_display_dpi_get (FALSE));

	layout = pango_layout_new (prend->pango_context);

	gog_debug (0, {
		char *msg = pango_font_description_to_string (
			prend->base.cur_style->font.font->desc);
		g_warning (msg);
		g_free (msg);
	});

	pango_layout_set_text (layout, text, -1);
	pango_layout_set_font_description (layout,
		prend->base.cur_style->font.font->desc);

	attr = pango_attr_scale_new (prend->base.zoom *
#warning COMPLETE CRAP.  This should not be necessary.  but scale seems to override size, rather than modify it
		pango_font_description_get_size (prend->base.cur_style->font.font->desc) /
		(PANGO_SCALE * 10)); /* assume base size == 10 */
	attr->start_index = 0;
	attr->end_index = -1;

	attrs = pango_attr_list_new ();
	pango_attr_list_insert (attrs, attr);

	pango_layout_set_attributes (layout, attrs);
	pango_attr_list_unref (attrs);

	return layout;
}

static void
gog_renderer_pixbuf_draw_text (GogRenderer *rend, ArtPoint const *pos, GtkAnchorType anchor,
			       char const *text, GogViewRequisition *size)
{
	FT_Bitmap ft_bitmap;
	GogRendererPixbuf *prend = GOG_RENDERER_PIXBUF (rend);
	PangoRectangle rect;
	PangoLayout   *layout = make_layout (prend, text);
	guint8 r, g, b, a, alpha, *dst, *src;
	int h, w, i;
	GogStyle const *style = rend->cur_style;
	ArtPoint real_pos = *pos;

	pango_layout_get_pixel_extents (layout, &rect, NULL);
	if (rect.width == 0 || rect.height == 0)
		return;
	ft_bitmap.rows         = rect.height;
	ft_bitmap.width        = rect.width;
	ft_bitmap.pitch        = (rect.width+3) & ~3;
	ft_bitmap.buffer       = g_malloc0 (ft_bitmap.rows * ft_bitmap.pitch);
	ft_bitmap.num_grays    = 256;
	ft_bitmap.pixel_mode   = ft_pixel_mode_grays;
	ft_bitmap.palette_mode = 0;
	ft_bitmap.palette      = NULL;
	pango_ft2_render_layout (&ft_bitmap, layout, -rect.x, -rect.y);

	r = UINT_RGBA_R (style->font.color);
	g = UINT_RGBA_G (style->font.color);
	b = UINT_RGBA_B (style->font.color);
	a = UINT_RGBA_A (style->font.color);

	switch (anchor) {
	case GTK_ANCHOR_CENTER : case GTK_ANCHOR_N : case GTK_ANCHOR_S :
		real_pos.x -= rect.width / 2;
		break;
	case GTK_ANCHOR_NE : case GTK_ANCHOR_SE : case GTK_ANCHOR_E :
		real_pos.x -= rect.width;
		break;
	default : break;
	}
	switch (anchor) {
	case GTK_ANCHOR_CENTER : case GTK_ANCHOR_E : case GTK_ANCHOR_W :
		real_pos.y -= rect.height / 2;
		break;
	case GTK_ANCHOR_SE : case GTK_ANCHOR_S : case GTK_ANCHOR_SW :
		real_pos.y -= rect.height;
		break;
	default : break;
	}

	/* do the compositing manually, ArtRender as used in librsvg is dog
	 * slow, and I do not feel like leaping through 20 different data
	 * structures to composite 1 byte images, onto rgba */
	dst = prend->pixels;
	dst += ((int)real_pos.y * prend->rowstride);
	dst += ((int)real_pos.x + rect.x)* 4;
	src = ft_bitmap.buffer;

	w = rect.width;
	if (size != NULL && w > size->w && size->w >= 0)
		w = size->w;

	h = rect.height;
	if (size != NULL && h > size->h && size->h >= 0)
		h = size->h;
	while (h--) {
		for (i = w; i-- > 0 ; dst += 4, src++) {
			/* FIXME: Do the libart thing instead of divide by 255 */
			alpha = (a * (*src)) / 255;
			dst[0] = (dst[0] * (255 - alpha) + r * alpha) / 255;
			dst[1] = (dst[1] * (255 - alpha) + g * alpha) / 255;
			dst[2] = (dst[2] * (255 - alpha) + b * alpha) / 255;
			dst[3] = (dst[3] * (255 - alpha) + a * alpha) / 255;
		}
		dst += prend->rowstride - w*4;
		src += ft_bitmap.pitch - w;
	}

	if (size != NULL) {
		pango_layout_get_pixel_extents (layout, &rect, NULL);
		size->w = rect.width;
		size->h = rect.height;
	}
	g_object_unref (layout);
}

static void
gog_renderer_pixbuf_draw_marker (GogRenderer *renderer, double x, double y)
{
	GdkRectangle r1, r2, dest; 
	GogStyle const *style = renderer->cur_style;
	GogRendererPixbuf *prend = GOG_RENDERER_PIXBUF (renderer);
	GdkPixbuf const *marker_pixbuf = go_marker_get_pixbuf (style->marker);

	if (marker_pixbuf == NULL)
		return;

	r2.x = r2.y = 0;
	r2.width = prend->w;
	r2.height = prend->h;

	r1.width = gdk_pixbuf_get_width (marker_pixbuf);
	r1.height = gdk_pixbuf_get_height (marker_pixbuf);
	r1.x = x - r1.width / 2;
	r1.y = y - r1.height / 2;

	if (gdk_rectangle_intersect (&r1, &r2, &dest))
		gdk_pixbuf_composite (marker_pixbuf, prend->buffer,
				      dest.x, dest.y,
				      dest.width, dest.height,
				      r1.x, r1.y,
				      1.0, 1.0,
				      GDK_INTERP_NEAREST,
				      255);
}

static void
gog_renderer_pixbuf_measure_text (GogRenderer *rend,
				  char const *text, GogViewRequisition *size)
{
	PangoRectangle  rect;
	PangoLayout    *layout = make_layout ((GogRendererPixbuf *)rend, text);
	pango_layout_get_pixel_extents (layout, &rect, NULL);
	g_object_unref (layout);

	size->w = rect.width;
	size->h = rect.height;
}

static void
gog_renderer_pixbuf_class_init (GogRendererClass *rend_klass)
{
	GObjectClass *gobject_klass   = (GObjectClass *) rend_klass;

	parent_klass = g_type_class_peek_parent (rend_klass);
	gobject_klass->finalize	  = gog_renderer_pixbuf_finalize;
	rend_klass->draw_path	  = gog_renderer_pixbuf_draw_path;
	rend_klass->draw_polygon  = gog_renderer_pixbuf_draw_polygon;
	rend_klass->draw_text	  = gog_renderer_pixbuf_draw_text;
	rend_klass->draw_marker	  = gog_renderer_pixbuf_draw_marker;
	rend_klass->measure_text  = gog_renderer_pixbuf_measure_text;
}

static void
gog_renderer_pixbuf_init (GogRendererPixbuf *prend)
{
	prend->buffer = NULL;
	prend->w = prend->h = 1; /* jsut in case */
}

GSF_CLASS (GogRendererPixbuf, gog_renderer_pixbuf,
	   gog_renderer_pixbuf_class_init, gog_renderer_pixbuf_init,
	   GOG_RENDERER_TYPE)

GdkPixbuf *
gog_renderer_pixbuf_get (GogRendererPixbuf *prend)
{
	g_return_val_if_fail (prend != NULL, NULL);

	return prend->buffer;
}

#if 0 /* An initial non-working attempt to use different dpi to render
	 different zooms */

/* fontmaps are reasonably expensive use a cache to share them */
static GHashTable *fontmap_cache = NULL; /* PangoFT2FontMap hashed by y_dpi */
static gboolean
cb_remove_entry (gpointer key, PangoFT2FontMap *value, PangoFT2FontMap *target)
{
	return value == target;
}
static void
cb_map_is_gone (gpointer data, GObject *where_the_object_was)
{
	g_warning ("fontmap %p is gone",where_the_object_was);
	g_hash_table_foreach_steal (fontmap_cache,
		(GHRFunc) cb_remove_entry, where_the_object_was);
}
static void
cb_weak_unref (GObject *fontmap)
{
	g_object_weak_unref (fontmap, cb_map_is_gone, NULL);
}
static PangoFT2FontMap *
fontmap_from_cache (double x_dpi, double y_dpi)
{
	PangoFT2FontMap *fontmap = NULL;
	int key_dpi = floor (y_dpi + .5);
	gpointer key = GUINT_TO_POINTER (key_dpi);

	if (fontmap_cache != NULL)
		fontmap = g_hash_table_lookup (fontmap_cache, key);
	else
		fontmap_cache = g_hash_table_new_full (g_direct_hash, g_direct_equal,
			NULL, (GDestroyNotify) cb_weak_unref);

	if (fontmap == NULL) {
		fontmap = PANGO_FT2_FONT_MAP (pango_ft2_font_map_new ());
		pango_ft2_font_map_set_resolution (fontmap, x_dpi, y_dpi);
		g_object_weak_ref (G_OBJECT (fontmap), cb_map_is_gone, NULL);
		g_hash_table_insert (fontmap_cache, key, fontmap);
	} else
		g_object_ref (fontmap);

	g_warning ("fontmap %d = %p", key_dpi, fontmap);
	return fontmap;
}
#endif

/**
 * gog_renderer_update :
 * @prend :
 * @w :
 * @h :
 *
 * Returns TRUE if the size actually changed.
 **/
gboolean
gog_renderer_pixbuf_update (GogRendererPixbuf *prend, int w, int h, double zoom)
{
	gboolean redraw = TRUE;
	GogView *view;
	GogViewAllocation allocation;

	g_return_val_if_fail (prend != NULL, FALSE);
	g_return_val_if_fail (prend->base.view != NULL, FALSE);

	view = prend->base.view;
	allocation.x = allocation.y = 0.;
	allocation.w = w;
	allocation.h = h;
	if (prend->w != w || prend->h != h) {
		double dpi_x, dpi_y;

		prend->w = w;
		prend->h = h;
		prend->base.scale_x = w / prend->base.logical_width_pts;
		prend->base.scale_y = h / prend->base.logical_height_pts;
		prend->base.scale = MIN (prend->base.scale_x, prend->base.scale_y);
		prend->base.zoom  = zoom;
		dpi_x = gog_renderer_pt2r_x (&prend->base, GO_IN_TO_PT (1.))
			/ zoom;
		dpi_y = gog_renderer_pt2r_y (&prend->base, GO_IN_TO_PT (1.))
			/ zoom;

		if (prend->buffer != NULL) {
			g_object_unref (prend->buffer);
			prend->buffer = NULL;
		}
		if (prend->pango_context != NULL) {
			g_object_unref (prend->pango_context);
			prend->pango_context = NULL;
		}

		prend->pango_context = pango_ft2_get_context (
			application_display_dpi_get (TRUE),
			application_display_dpi_get (FALSE));

		/* make sure we dont try to queue an update while updating */
		prend->base.needs_update = TRUE;

		/* scale just changed need to recalculate sizes */
		gog_renderer_invalidate_size_requests (&prend->base);
		gog_view_size_allocate (view, &allocation);
	} else if (w != view->allocation.w || h != view->allocation.h)
		gog_view_size_allocate (view, &allocation);
	else
		redraw = gog_view_update_sizes (view);

	redraw |= prend->base.needs_update;
	prend->base.needs_update = FALSE;

	gog_debug (0, g_warning ("rend_pixbuf:update = %d", redraw););

	if (redraw) {
		if (prend->buffer == NULL) {
			prend->buffer = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8,
							prend->w, prend->h);
			prend->pixels    = gdk_pixbuf_get_pixels (prend->buffer);
			prend->rowstride = gdk_pixbuf_get_rowstride (prend->buffer);
		}
		gdk_pixbuf_fill (prend->buffer, 0);

		gog_view_render	(view, NULL);
	}

	return redraw;
}
